#include "ehs/io/Console_W32.h"

namespace ehs
{
	void* Console::hdlOut = nullptr;
	void* Console::hdlIn = nullptr;
	bool Console::isConsole = true;

	void Console::Attach()
	{
		if (!AttachConsole(ATTACH_PARENT_PROCESS))
		{
			DWORD code = GetLastError();
			if (code == ERROR_INVALID_HANDLE)
				return;
		}

		hdlIn = GetStdHandle(STD_INPUT_HANDLE);
		hdlOut = GetStdHandle(STD_OUTPUT_HANDLE);

		DWORD mode = 0;

		if (!GetConsoleMode(hdlOut, &mode))
		{
			if (GetLastError() == ERROR_INVALID_HANDLE)
				isConsole = false;
		}
		else
		{
			//DWORD code = WaitForSingleObject(hdlIn, 15000);
			//if (code == WAIT_FAILED || code == WAIT_TIMEOUT || code == WAIT_ABANDONED)
			//	EHS_LOG_INT("Error", 0, "Failed to wait for console input.");

			isConsole = true;
		}
	}

	void Console::Free()
	{
		hdlOut = nullptr;
		hdlIn = nullptr;
	}

	void Console::Write_32(const Str_32& str, const bool newLine)
	{
		if (isConsole)
		{
			Str_16 r = UTF::To_16(str);
			if (newLine)
				r += L"\r\n";

			DWORD offset = 0;

			do
			{
				DWORD written = 0;
				if (!WriteConsoleW(hdlOut, &r[offset], (DWORD)r.Size() - offset, &written, nullptr))
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

				offset += written;
			}
			while (offset < r.Size());
		}
		else
		{
			Str_8 r = UTF::To_8(str);
			if (newLine)
				r += "\r\n";

			DWORD offset = 0;

			do
			{
				DWORD written = 0;
				if (!WriteFile(hdlOut, &((Char_8*)r)[offset], (DWORD)r.Size(true) - offset, &written, nullptr))
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

				offset += written;
			}
			while (offset < r.Size(true));
		}
	}

	void Console::Write_16(const Str_16& str, const bool newLine)
	{
		if (isConsole)
		{
			Str_16 r = str;
			if (newLine)
				r += L"\r\n";

			DWORD offset = 0;

			do
			{
				DWORD written = 0;
				if (!WriteConsoleW(hdlOut, &r[offset], (DWORD)r.Size() - offset, &written, nullptr))
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

				offset += written;
			}
			while (offset < r.Size());
		}
		else
		{
			Str_8 r = UTF::To_8(str);
			if (newLine)
				r += "\r\n";

			DWORD offset = 0;

			do
			{
				DWORD written = 0;
				if (!WriteFile(hdlOut, &((Char_8*)r)[offset], (DWORD)r.Size(true) - offset, &written, nullptr))
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

				offset += written;
			}
			while (offset < r.Size(true));
		}
	}

	void Console::Write_8(const Str_8& str, const bool newLine)
	{
		if (isConsole)
		{
			Str_16 r = UTF::To_16(str);
			if (newLine)
				r += L"\r\n";

			DWORD offset = 0;

			do
			{
				DWORD written = 0;
				if (!WriteConsoleW(hdlOut, &r[offset], (DWORD)r.Size() - offset, &written, nullptr))
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

				offset += written;
			}
			while (offset < r.Size());
		}
		else
		{
			Str_8 r = str;
			if (newLine)
				r += "\r\n";

			DWORD offset = 0;

			do
			{
				DWORD written = 0;
				if (!WriteFile(hdlOut, &r[offset], (DWORD)r.Size() - offset, &written, nullptr))
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

				offset += written;
			}
			while (offset < r.Size());
		}
	}

	Str_32 Console::Read_32(const UInt_64 bufferSize)
	{
		if (isConsole)
		{
			Str_16 result;

			DWORD offset = 0;

			do
			{
				result.Resize(result.Size() + bufferSize);

				DWORD read = 0;

				if (!ReadConsoleW(hdlIn, &result[offset], (DWORD)bufferSize, &read, nullptr))
				{
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

					return U"";
				}

				offset += read;
			}
			while (result[offset - 1] != L'\n');

			if (offset >= 2 && result[offset - 2] == L'\r' && result[offset - 1] == L'\n')
				result.Resize(offset - 2);
			else
				result.Resize(offset - 1);

			return UTF::To_32(result);
		}
		else
		{
			Str_8 result;

			DWORD offset = 0;

			do
			{
				result.Resize(result.Size() + bufferSize);

				DWORD read = 0;

				if (!ReadFile(hdlIn, &result[offset], (DWORD)bufferSize, &read, nullptr))
				{
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

					return U"";
				}

				offset += read;
			}
			while (result[offset - 1] != '\n');

			if (offset >= 2 && result[offset - 2] == '\r' && result[offset - 1] == '\n')
				result.Resize(offset - 2);
			else
				result.Resize(offset - 1);

			return UTF::To_32(result);
		}
	}

	Str_16 Console::Read_16(const UInt_64 bufferSize)
	{
		if (isConsole)
		{
			Str_16 result;

			DWORD offset = 0;

			do
			{
				result.Resize(result.Size() + bufferSize);

				DWORD read = 0;

				if (!ReadConsoleW(hdlIn, &result[offset], (DWORD)bufferSize, &read, nullptr))
				{
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

					return L"";
				}

				offset += read;
			}
			while (result[offset - 1] != L'\n');

			if (offset >= 2 && result[offset - 2] == L'\r' && result[offset - 1] == L'\n')
				result.Resize(offset - 2);
			else
				result.Resize(offset - 1);

			return result;
		}
		else
		{
			Str_8 result;

			DWORD offset = 0;

			do
			{
				result.Resize(result.Size() + bufferSize);

				DWORD read = 0;

				if (!ReadFile(hdlIn, &result[offset], (DWORD)bufferSize, &read, nullptr))
				{
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

					return L"";
				}

				offset += read;
			}
			while (result[offset - 1] != '\n');

			if (offset >= 2 && result[offset - 2] == '\r' && result[offset - 1] == '\n')
				result.Resize(offset - 2);
			else
				result.Resize(offset - 1);

			return UTF::To_16(result);
		}
	}

	Str_8 Console::Read_8(const UInt_64 bufferSize)
	{
		if (isConsole)
		{
			Str_16 result;

			DWORD offset = 0;

			do
			{
				result.Resize(result.Size() + bufferSize);

				DWORD read = 0;

				if (!ReadConsoleW(hdlIn, &result[offset], (DWORD)bufferSize, &read, nullptr))
				{
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

					return "";
				}

				offset += read;
			}
			while (result[offset - 1] != L'\n');

			if (offset >= 2 && result[offset - 2] == L'\r' && result[offset - 1] == L'\n')
				result.Resize(offset - 2);
			else
				result.Resize(offset - 1);

			return UTF::To_8(result);
		}
		else
		{
			Str_8 result;

			DWORD offset = 0;

			do
			{
				result.Resize(result.Size() + bufferSize);

				DWORD read = 0;

				if (!ReadFile(hdlIn, &result[offset], (DWORD)bufferSize, &read, nullptr))
				{
					EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");

					return "";
				}

				offset += read;
			}
			while (result[offset - 1] != '\n');

			if (offset >= 2 && result[offset - 2] == '\r' && result[offset - 1] == '\n')
				result.Resize(offset - 2);
			else
				result.Resize(offset - 1);

			return result;
		}
	}

	void Console::Clear()
	{
		CONSOLE_SCREEN_BUFFER_INFO info = {};
		if (!GetConsoleScreenBufferInfo(hdlOut, &info))
			return;

		DWORD size = info.dwSize.X * info.dwSize.Y;

		DWORD written = 0;
		if (!FillConsoleOutputCharacterW(hdlOut, L' ', size, {0, 0}, &written))
			return;

		if (!GetConsoleScreenBufferInfo(hdlOut, &info))
			return;

		if (!FillConsoleOutputAttribute(hdlOut, info.wAttributes, size, {0, 0}, &written))
			return;

		SetConsoleCursorPosition(hdlOut, {0, 0});
	}

	void Console::SetTitle_32(const Str_32& title)
	{
		if (!SetConsoleTitleW(UTF::To_16(title)))
			EHS_LOG_INT("Error", 0, "Failed to set console title with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	void Console::SetTitle_16(const Str_16& title)
	{
		if (!SetConsoleTitleW(title))
			EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	void Console::SetTitle_8(const Str_8& title)
	{
		if (!SetConsoleTitleW(UTF::To_16(title)))
			EHS_LOG_INT("Error", 0, "Failed with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	Str_32 Console::GetTitle_32()
	{
		Str_16 title(EHS_MAX_PATH);

		DWORD size = GetConsoleTitleW(&title[0], EHS_MAX_PATH);

		title.Resize(size);

		return UTF::To_32(title);
	}

	Str_16 Console::GetTitle_16()
	{
		Str_16 title(EHS_MAX_PATH);

		DWORD size = GetConsoleTitleW(&title[0], EHS_MAX_PATH);

		title.Resize(size);

		return title;
	}

	Str_8 Console::GetTitle_8()
	{
		Str_16 title(EHS_MAX_PATH);

		DWORD size = GetConsoleTitleW(&title[0], EHS_MAX_PATH);

		title.Resize(size);

		return UTF::To_8(title);
	}

	Vector<Str_32> Console::GetArgs_32(const UInt_64 bufferSize)
	{
		return UTF::To_32(GetCommandLineW()).Split(U" ");
	}

	Vector<Str_16> Console::GetArgs_16(const UInt_64 bufferSize)
	{
		return Str_16(GetCommandLineW()).Split(L" ");
	}

	Vector<Str_8> Console::GetArgs_8(const UInt_64 bufferSize)
	{
		return UTF::To_8(GetCommandLineW()).Split(" ");
	}
}