#include "ehs/system/BaseThread.h"
#include "ehs/system/CPU.h"

namespace ehs
{
	BaseThread::BaseThread()
		: stackSize(0), id(0)
	{
	}

	BaseThread::BaseThread(const UInt_64 stackSize)
		: stackSize(stackSize), id(0)
	{
	}

	BaseThread::BaseThread(BaseThread&& thread) noexcept
		: stackSize(thread.stackSize), id(thread.id)
	{
		thread.stackSize = 0;
		thread.id = 0;
	}

	BaseThread::BaseThread(const BaseThread& thread)
		: stackSize(thread.stackSize), id(0)
	{
	}

	BaseThread& BaseThread::operator=(BaseThread&& thread) noexcept
	{
		if (this == &thread)
			return *this;

		stackSize = thread.stackSize;
		id = thread.id;

		thread.stackSize = 0;
		thread.id = 0;

		return *this;
	}

	BaseThread& BaseThread::operator=(const BaseThread& thread)
	{
		if (this == &thread)
			return *this;

		stackSize = thread.stackSize;
		id = 0;

		return *this;
	}

	void BaseThread::Detach()
	{
		stackSize = 0;
		id = 0;
	}

	UInt_64 BaseThread::GetStackSize() const
	{
		return stackSize;
	}

	UInt_32 BaseThread::GetId() const
	{
		return id;
	}

	UInt_32 BaseThread::GetCurrentId()
	{
		return 0;
	}

	bool BaseThread::IsValid() const
	{
		return stackSize && id;
	}

	float BaseThread::SoftSleep(const float seconds)
	{
		UInt_64 freq = CPU::GetTSC_Freq();
		UInt_64 start = CPU::GetTSC();

		float elapsed = 0.0f;
		while ((elapsed = (float)(CPU::GetTSC() - start) / (float)freq) <= seconds);

		return elapsed;
	}

	void BaseThread::HardSleep(const UInt_32 milliseconds)
	{
	}
}